from typing import Dict

from yfinance import utils
from yfinance.data import YfData
from yfinance.const import _BASE_URL_, PREDEFINED_SCREENER_BODY_MAP
from .screener_query import Query

_SCREENER_URL_ = f"{_BASE_URL_}/v1/finance/screener"

class Screener:
    def __init__(self, session=None, proxy=None):
        self.proxy = proxy
        self.session = session

        self._data: YfData = YfData(session=session)
        self._body: Dict = {}
        self._response: Dict = {}
        self._body_updated = False
        self._accepted_body_keys = {"offset","size","sortField","sortType","quoteType","query","userId","userIdType"}
        self._predefined_bodies = PREDEFINED_SCREENER_BODY_MAP.keys()

    @property
    def body(self) -> Dict:
        return self._body
    
    @property
    def response(self) -> Dict:
        if self._body_updated or self._response is None:
            self._fetch_and_parse()
        
        self._body_updated = False
        return self._response
    
    @property
    def predefined_bodies(self) -> Dict:
        return self._predefined_bodies

    def set_default_body(self, query: Query, offset: int = 0, size: int = 100, sortField: str = "ticker", sortType: str = "desc", quoteType: str = "equity", userId: str = "", userIdType: str = "guid") -> None:
        self._body_updated = True

        self._body = {
            "offset": offset,
            "size": size,
            "sortField": sortField,
            "sortType": sortType,
            "quoteType": quoteType,
            "query": query.to_dict(),
            "userId": userId,
            "userIdType": userIdType
        }

    def set_predefined_body(self, k: str) -> None:
        body = PREDEFINED_SCREENER_BODY_MAP.get(k, None)
        if not body:
            raise ValueError(f'Invalid key {k} provided for predefined screener')
        
        self._body_updated = True
        self._body = body

    def set_body(self, body: Dict) -> None:
        missing_keys = [key for key in self._accepted_body_keys if key not in body]
        if missing_keys:
            raise ValueError(f"Missing required keys in body: {missing_keys}")

        extra_keys = [key for key in body if key not in self._accepted_body_keys]
        if extra_keys:
            raise ValueError(f"Body contains extra keys: {extra_keys}")

        self._body_updated = True
        self._body = body


    def patch_body(self, values: Dict) -> None:
        extra_keys = [key for key in values if key not in self._accepted_body_keys]
        if extra_keys:
            raise ValueError(f"Body contains extra keys: {extra_keys}")
        
        self._body_updated = True
        for k in values:
            self._body[k] = values[k]

    def _validate_body(self) -> None:
        if not all(k in self._body for k in self._accepted_body_keys):
            raise ValueError("Missing required keys in body")

    def _fetch(self) -> Dict:
        params_dict = {"corsDomain": "finance.yahoo.com", "formatted": "false", "lang": "en-US", "region": "US"}
        response = self._data.post(_SCREENER_URL_, body=self.body, user_agent_headers=self._data.user_agent_headers, params=params_dict, proxy=self.proxy)
        response.raise_for_status()
        return response.json()
    
    def _fetch_and_parse(self) -> None:
        response = None
        self._validate_body()
        
        try:
            response = self._fetch()
            self._response = response['finance']['result'][0]
        except Exception as e:
            logger = utils.get_yf_logger()
            logger.error(f"Failed to get screener data for '{self._body.get('query', 'query not set')}' reason: {e}")
            logger.debug("Got response: ")
            logger.debug("-------------")
            logger.debug(f" {response}")
            logger.debug("-------------")



import pandas as pd
import requests

from yfinance import utils
from yfinance.data import YfData
from yfinance.const import quote_summary_valid_modules
from yfinance.scrapers.quote import _QUOTE_SUMMARY_URL_
from yfinance.exceptions import YFException


class Analysis:

    def __init__(self, data: YfData, symbol: str, proxy=None):
        self._data = data
        self._symbol = symbol
        self.proxy = proxy

        # In quoteSummary the 'earningsTrend' module contains most of the data below.
        # The format of data is not optimal so each function will process it's part of the data.
        # This variable works as a cache.
        self._earnings_trend = None

        self._analyst_price_targets = None
        self._earnings_estimate = None
        self._revenue_estimate = None
        self._earnings_history = None
        self._eps_trend = None
        self._eps_revisions = None
        self._growth_estimates = None

    @property
    def analyst_price_targets(self) -> dict:
        if self._analyst_price_targets is not None:
            return self._analyst_price_targets

        try:
            data = self._fetch(['financialData'])
            data = data['quoteSummary']['result'][0]['financialData']
        except (TypeError, KeyError):
            self._analyst_price_targets = {}
            return self._analyst_price_targets

        keys = [
            ('currentPrice', 'current'),
            ('targetLowPrice', 'low'),
            ('targetHighPrice', 'high'),
            ('targetMeanPrice', 'mean'),
            ('targetMedianPrice', 'median'),
        ]

        self._analyst_price_targets = {newKey: data.get(oldKey, None) for oldKey, newKey in keys}
        return self._analyst_price_targets

    @property
    def earnings_estimate(self) -> pd.DataFrame:
        if self._earnings_estimate is not None:
            return self._earnings_estimate

        if self._earnings_trend is None:
            self._fetch_earnings_trend()

        data_dict = {
            'numberOfAnalysts': [],
            'avg': [],
            'low': [],
            'high': [],
            'yearAgoEps': [],
            'growth': []
        }
        periods = []

        for item in self._earnings_trend[:4]:
            periods.append(item['period'])
            earnings_estimate = item.get('earningsEstimate', {})

            for key in data_dict.keys():
                data_dict[key].append(earnings_estimate.get(key, {}).get('raw', None))

        self._earnings_estimate = pd.DataFrame(data_dict, index=periods)
        return self._earnings_estimate

    @property
    def revenue_estimate(self) -> pd.DataFrame:
        if self._revenue_estimate is not None:
            return self._revenue_estimate

        if self._earnings_trend is None:
            self._fetch_earnings_trend()

        data_dict = {
            'numberOfAnalysts': [],
            'avg': [],
            'low': [],
            'high': [],
            'yearAgoRevenue': [],
            'growth': []
        }
        periods = []

        for item in self._earnings_trend[:4]:
            periods.append(item['period'])
            revenue_estimate = item.get('revenueEstimate', {})

            for key in data_dict.keys():
                data_dict[key].append(revenue_estimate.get(key, {}).get('raw', None))

        self._revenue_estimate = pd.DataFrame(data_dict, index=periods)
        return self._revenue_estimate

    @property
    def earnings_history(self) -> pd.DataFrame:
        if self._earnings_history is not None:
            return self._earnings_history

        try:
            data = self._fetch(['earningsHistory'])
            data = data['quoteSummary']['result'][0]['earningsHistory']['history']
        except (TypeError, KeyError):
            self._earnings_history = pd.DataFrame()
            return self._earnings_history

        data_dict = {
            'epsEstimate': [],
            'epsActual': [],
            'epsDifference': [],
            'surprisePercent': []
        }
        quarters = []

        for item in data:
            quarters.append(item.get('quarter', {}).get('fmt', None))

            for key in data_dict.keys():
                data_dict[key].append(item.get(key, {}).get('raw', None))
        
        datetime_index = pd.to_datetime(quarters, format='%Y-%m-%d')
        self._earnings_history = pd.DataFrame(data_dict, index=datetime_index)
        return self._earnings_history

    @property
    def eps_trend(self) -> pd.DataFrame:
        if self._eps_trend is not None:
            return self._eps_trend

        if self._earnings_trend is None:
            self._fetch_earnings_trend()

        data_dict = {
            'current': [],
            '7daysAgo': [],
            '30daysAgo': [],
            '60daysAgo': [],
            '90daysAgo': []
        }
        periods = []

        for item in self._earnings_trend[:4]:
            periods.append(item['period'])
            eps_trend = item.get('epsTrend', {})

            for key in data_dict.keys():
                data_dict[key].append(eps_trend.get(key, {}).get('raw', None))

        self._eps_trend = pd.DataFrame(data_dict, index=periods)
        return self._eps_trend

    @property
    def eps_revisions(self) -> pd.DataFrame:
        if self._eps_revisions is not None:
            return self._eps_revisions

        if self._earnings_trend is None:
            self._fetch_earnings_trend()

        data_dict = {
            'upLast7days': [],
            'upLast30days': [],
            'downLast7days': [],
            'downLast30days': []
        }
        periods = []

        for item in self._earnings_trend[:4]:
            periods.append(item['period'])
            eps_revisions = item.get('epsRevisions', {})

            for key in data_dict.keys():
                data_dict[key].append(eps_revisions.get(key, {}).get('raw', None))

        self._eps_revisions = pd.DataFrame(data_dict, index=periods)
        return self._eps_revisions

    @property
    def growth_estimates(self) -> pd.DataFrame:
        if self._growth_estimates is not None:
            return self._growth_estimates

        if self._earnings_trend is None:
            self._fetch_earnings_trend()

        try:
            trends = self._fetch(['industryTrend', 'sectorTrend', 'indexTrend'])
            trends = trends['quoteSummary']['result'][0]
        except (TypeError, KeyError):
            self._growth_estimates = pd.DataFrame()
            return self._growth_estimates

        data_dict = {
            '0q': [],
            '+1q': [],
            '0y': [],
            '+1y': [],
            '+5y': [],
            '-5y': []
        }

        # make sure no column is empty
        dummy_trend = [{'period': key, 'growth': None} for key in data_dict.keys()]
        industry_trend = trends['industryTrend']['estimates'] or dummy_trend
        sector_trend = trends['sectorTrend']['estimates'] or dummy_trend
        index_trend = trends['indexTrend']['estimates'] or dummy_trend

        for item in self._earnings_trend:
            period = item['period']
            data_dict[period].append(item.get('growth', {}).get('raw', None))

        for item in industry_trend:
            period = item['period']
            data_dict[period].append(item.get('growth', None))

        for item in sector_trend:
            period = item['period']
            data_dict[period].append(item.get('growth', None))

        for item in index_trend:
            period = item['period']
            data_dict[period].append(item.get('growth', None))

        cols = ['stock', 'industry', 'sector', 'index']
        self._growth_estimates = pd.DataFrame(data_dict, index=cols).T
        return self._growth_estimates

    # modified version from quote.py
    def _fetch(self, modules: list):
        if not isinstance(modules, list):
            raise YFException("Should provide a list of modules, see available modules using `valid_modules`")

        modules = ','.join([m for m in modules if m in quote_summary_valid_modules])
        if len(modules) == 0:
            raise YFException("No valid modules provided, see available modules using `valid_modules`")
        params_dict = {"modules": modules, "corsDomain": "finance.yahoo.com", "formatted": "false", "symbol": self._symbol}
        try:
            result = self._data.get_raw_json(_QUOTE_SUMMARY_URL_ + f"/{self._symbol}", user_agent_headers=self._data.user_agent_headers, params=params_dict, proxy=self.proxy)
        except requests.exceptions.HTTPError as e:
            utils.get_yf_logger().error(str(e))
            return None
        return result

    def _fetch_earnings_trend(self) -> None:
        try:
            data = self._fetch(['earningsTrend'])
            self._earnings_trend = data['quoteSummary']['result'][0]['earningsTrend']['trend']
        except (TypeError, KeyError):
            self._earnings_trend = []



"""
Tests for utils

To run all tests in suite from commandline:
   python -m unittest tests.utils

Specific test class:
   python -m unittest tests.utils.TestTicker

"""
from datetime import datetime
from unittest import TestSuite

import pandas as pd
# import numpy as np

from tests.context import yfinance as yf

import unittest
# import requests_cache
import tempfile
import os


class TestCache(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tempCacheDir = tempfile.TemporaryDirectory()
        yf.set_tz_cache_location(cls.tempCacheDir.name)

    @classmethod
    def tearDownClass(cls):
        cls.tempCacheDir.cleanup()

    def test_storeTzNoRaise(self):
        # storing TZ to cache should never raise exception
        tkr = 'AMZN'
        tz1 = "America/New_York"
        tz2 = "London/Europe"
        cache = yf.cache.get_tz_cache()
        cache.store(tkr, tz1)
        cache.store(tkr, tz2)

    def test_setTzCacheLocation(self):
        self.assertEqual(yf.cache._TzDBManager.get_location(), self.tempCacheDir.name)

        tkr = 'AMZN'
        tz1 = "America/New_York"
        cache = yf.cache.get_tz_cache()
        cache.store(tkr, tz1)

        self.assertTrue(os.path.exists(os.path.join(self.tempCacheDir.name, "tkr-tz.db")))


class TestCacheNoPermission(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        yf.set_tz_cache_location("/root/yf-cache")

    def test_tzCacheRootStore(self):
        # Test that if cache path in read-only filesystem, no exception.
        tkr = 'AMZN'
        tz1 = "America/New_York"

        # During attempt to store, will discover cannot write
        yf.cache.get_tz_cache().store(tkr, tz1)

        # Handling the store failure replaces cache with a dummy
        cache = yf.cache.get_tz_cache()
        self.assertTrue(cache.dummy)
        cache.store(tkr, tz1)

    def test_tzCacheRootLookup(self):
        # Test that if cache path in read-only filesystem, no exception.
        tkr = 'AMZN'
        # During attempt to lookup, will discover cannot write
        yf.cache.get_tz_cache().lookup(tkr)

        # Handling the lookup failure replaces cache with a dummy
        cache = yf.cache.get_tz_cache()
        self.assertTrue(cache.dummy)
        cache.lookup(tkr)


class TestPandas(unittest.TestCase):
    date_strings = ["2024-08-07 09:05:00+02:00", "2024-08-07 09:05:00-04:00"]

    @unittest.expectedFailure
    def test_mixed_timezones_to_datetime_fails(self):
        series = pd.Series(self.date_strings)
        series = series.map(pd.Timestamp)
        converted = pd.to_datetime(series)
        self.assertIsNotNone(converted[0].tz)

    def test_mixed_timezones_to_datetime(self):
        series = pd.Series(self.date_strings)
        series = series.map(pd.Timestamp)
        converted = pd.to_datetime(series, utc=True)
        self.assertIsNotNone(converted[0].tz)
        i = 0
        for dt in converted:
            dt: datetime
            ts: pd.Timestamp = series[i]
            self.assertEqual(dt.isoformat(), ts.tz_convert(tz="UTC").isoformat())
            i += 1


def suite():
    ts: TestSuite = unittest.TestSuite()
    ts.addTest(TestCache('Test cache'))
    ts.addTest(TestCacheNoPermission('Test cache no permission'))
    ts.addTest(TestPandas("Test pandas"))
    return ts


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock
from yfinance.const import PREDEFINED_SCREENER_BODY_MAP
from yfinance.screener.screener import Screener
from yfinance.screener.screener_query import EquityQuery


class TestScreener(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.screener = Screener()
        self.query = EquityQuery('gt',['eodprice',3])

    def test_set_default_body(self):
        self.screener.set_default_body(self.query)

        self.assertEqual(self.screener.body['offset'], 0)
        self.assertEqual(self.screener.body['size'], 100)
        self.assertEqual(self.screener.body['sortField'], 'ticker')
        self.assertEqual(self.screener.body['sortType'], 'desc')
        self.assertEqual(self.screener.body['quoteType'], 'equity')
        self.assertEqual(self.screener.body['query'], self.query.to_dict())
        self.assertEqual(self.screener.body['userId'], '')
        self.assertEqual(self.screener.body['userIdType'], 'guid')

    def test_set_predefined_body(self):
        k = 'most_actives'
        self.screener.set_predefined_body(k)
        self.assertEqual(self.screener.body, PREDEFINED_SCREENER_BODY_MAP[k])

    def test_set_predefined_body_invalid_key(self):
        with self.assertRaises(ValueError):
            self.screener.set_predefined_body('invalid_key')

    def test_set_body(self):
        body = {
            "offset": 0,
            "size": 100,
            "sortField": "ticker",
            "sortType": "desc",
            "quoteType": "equity",
            "query": self.query.to_dict(),
            "userId": "",
            "userIdType": "guid"
        }
        self.screener.set_body(body)

        self.assertEqual(self.screener.body, body)

    def test_set_body_missing_keys(self):
        body = {
            "offset": 0,
            "size": 100,
            "sortField": "ticker",
            "sortType": "desc",
            "quoteType": "equity"
        }
        with self.assertRaises(ValueError):
            self.screener.set_body(body)

    def test_set_body_extra_keys(self):
        body = {
            "offset": 0,
            "size": 100,
            "sortField": "ticker",
            "sortType": "desc",
            "quoteType": "equity",
            "query": self.query.to_dict(),
            "userId": "",
            "userIdType": "guid",
            "extraKey": "extraValue"
        }
        with self.assertRaises(ValueError):
            self.screener.set_body(body)

    def test_patch_body(self):
        initial_body = {
            "offset": 0,
            "size": 100,
            "sortField": "ticker",
            "sortType": "desc",
            "quoteType": "equity",
            "query": self.query.to_dict(),
            "userId": "",
            "userIdType": "guid"
        }
        self.screener.set_body(initial_body)
        patch_values = {"size": 50}
        self.screener.patch_body(patch_values)

        self.assertEqual(self.screener.body['size'], 50)
        self.assertEqual(self.screener.body['query'], self.query.to_dict())

    def test_patch_body_extra_keys(self):
        initial_body = {
            "offset": 0,
            "size": 100,
            "sortField": "ticker",
            "sortType": "desc",
            "quoteType": "equity",
            "query": self.query.to_dict(),
            "userId": "",
            "userIdType": "guid"
        }
        self.screener.set_body(initial_body)
        patch_values = {"extraKey": "extraValue"}
        with self.assertRaises(ValueError):
            self.screener.patch_body(patch_values)

    @patch('yfinance.screener.screener.YfData.post')
    def test_fetch(self, mock_post):
        mock_response = MagicMock()
        mock_response.json.return_value = {'finance': {'result': [{}]}}
        mock_post.return_value = mock_response

        self.screener.set_default_body(self.query)
        response = self.screener._fetch()

        self.assertEqual(response, {'finance': {'result': [{}]}})

    @patch('yfinance.screener.screener.YfData.post')
    def test_fetch_and_parse(self, mock_post):
        mock_response = MagicMock()
        mock_response.json.return_value = {'finance': {'result': [{'key': 'value'}]}}
        mock_post.return_value = mock_response

        self.screener.set_default_body(self.query)
        self.screener._fetch_and_parse()
        self.assertEqual(self.screener.response, {'key': 'value'})

if __name__ == '__main__':
    unittest.main()


_QUERY1_URL_ = 'https://query1.finance.yahoo.com'
_BASE_URL_ = 'https://query2.finance.yahoo.com'
_ROOT_URL_ = 'https://finance.yahoo.com'

fundamentals_keys = {
    'financials': ["TaxEffectOfUnusualItems", "TaxRateForCalcs", "NormalizedEBITDA", "NormalizedDilutedEPS",
                   "NormalizedBasicEPS", "TotalUnusualItems", "TotalUnusualItemsExcludingGoodwill",
                   "NetIncomeFromContinuingOperationNetMinorityInterest", "ReconciledDepreciation",
                   "ReconciledCostOfRevenue", "EBITDA", "EBIT", "NetInterestIncome", "InterestExpense",
                   "InterestIncome", "ContinuingAndDiscontinuedDilutedEPS", "ContinuingAndDiscontinuedBasicEPS",
                   "NormalizedIncome", "NetIncomeFromContinuingAndDiscontinuedOperation", "TotalExpenses",
                   "RentExpenseSupplemental", "ReportedNormalizedDilutedEPS", "ReportedNormalizedBasicEPS",
                   "TotalOperatingIncomeAsReported", "DividendPerShare", "DilutedAverageShares", "BasicAverageShares",
                   "DilutedEPS", "DilutedEPSOtherGainsLosses", "TaxLossCarryforwardDilutedEPS",
                   "DilutedAccountingChange", "DilutedExtraordinary", "DilutedDiscontinuousOperations",
                   "DilutedContinuousOperations", "BasicEPS", "BasicEPSOtherGainsLosses", "TaxLossCarryforwardBasicEPS",
                   "BasicAccountingChange", "BasicExtraordinary", "BasicDiscontinuousOperations",
                   "BasicContinuousOperations", "DilutedNIAvailtoComStockholders", "AverageDilutionEarnings",
                   "NetIncomeCommonStockholders", "OtherunderPreferredStockDividend", "PreferredStockDividends",
                   "NetIncome", "MinorityInterests", "NetIncomeIncludingNoncontrollingInterests",
                   "NetIncomeFromTaxLossCarryforward", "NetIncomeExtraordinary", "NetIncomeDiscontinuousOperations",
                   "NetIncomeContinuousOperations", "EarningsFromEquityInterestNetOfTax", "TaxProvision",
                   "PretaxIncome", "OtherIncomeExpense", "OtherNonOperatingIncomeExpenses", "SpecialIncomeCharges",
                   "GainOnSaleOfPPE", "GainOnSaleOfBusiness", "OtherSpecialCharges", "WriteOff",
                   "ImpairmentOfCapitalAssets", "RestructuringAndMergernAcquisition", "SecuritiesAmortization",
                   "EarningsFromEquityInterest", "GainOnSaleOfSecurity", "NetNonOperatingInterestIncomeExpense",
                   "TotalOtherFinanceCost", "InterestExpenseNonOperating", "InterestIncomeNonOperating",
                   "OperatingIncome", "OperatingExpense", "OtherOperatingExpenses", "OtherTaxes",
                   "ProvisionForDoubtfulAccounts", "DepreciationAmortizationDepletionIncomeStatement",
                   "DepletionIncomeStatement", "DepreciationAndAmortizationInIncomeStatement", "Amortization",
                   "AmortizationOfIntangiblesIncomeStatement", "DepreciationIncomeStatement", "ResearchAndDevelopment",
                   "SellingGeneralAndAdministration", "SellingAndMarketingExpense", "GeneralAndAdministrativeExpense",
                   "OtherGandA", "InsuranceAndClaims", "RentAndLandingFees", "SalariesAndWages", "GrossProfit",
                   "CostOfRevenue", "TotalRevenue", "ExciseTaxes", "OperatingRevenue", "LossAdjustmentExpense",
                   "NetPolicyholderBenefitsAndClaims", "PolicyholderBenefitsGross", "PolicyholderBenefitsCeded",
                   "OccupancyAndEquipment", "ProfessionalExpenseAndContractServicesExpense", "OtherNonInterestExpense"],
    'balance-sheet': ["TreasurySharesNumber", "PreferredSharesNumber", "OrdinarySharesNumber", "ShareIssued", "NetDebt",
                      "TotalDebt", "TangibleBookValue", "InvestedCapital", "WorkingCapital", "NetTangibleAssets",
                      "CapitalLeaseObligations", "CommonStockEquity", "PreferredStockEquity", "TotalCapitalization",
                      "TotalEquityGrossMinorityInterest", "MinorityInterest", "StockholdersEquity",
                      "OtherEquityInterest", "GainsLossesNotAffectingRetainedEarnings", "OtherEquityAdjustments",
                      "FixedAssetsRevaluationReserve", "ForeignCurrencyTranslationAdjustments",
                      "MinimumPensionLiabilities", "UnrealizedGainLoss", "TreasuryStock", "RetainedEarnings",
                      "AdditionalPaidInCapital", "CapitalStock", "OtherCapitalStock", "CommonStock", "PreferredStock",
                      "TotalPartnershipCapital", "GeneralPartnershipCapital", "LimitedPartnershipCapital",
                      "TotalLiabilitiesNetMinorityInterest", "TotalNonCurrentLiabilitiesNetMinorityInterest",
                      "OtherNonCurrentLiabilities", "LiabilitiesHeldforSaleNonCurrent", "RestrictedCommonStock",
                      "PreferredSecuritiesOutsideStockEquity", "DerivativeProductLiabilities", "EmployeeBenefits",
                      "NonCurrentPensionAndOtherPostretirementBenefitPlans", "NonCurrentAccruedExpenses",
                      "DuetoRelatedPartiesNonCurrent", "TradeandOtherPayablesNonCurrent",
                      "NonCurrentDeferredLiabilities", "NonCurrentDeferredRevenue",
                      "NonCurrentDeferredTaxesLiabilities", "LongTermDebtAndCapitalLeaseObligation",
                      "LongTermCapitalLeaseObligation", "LongTermDebt", "LongTermProvisions", "CurrentLiabilities",
                      "OtherCurrentLiabilities", "CurrentDeferredLiabilities", "CurrentDeferredRevenue",
                      "CurrentDeferredTaxesLiabilities", "CurrentDebtAndCapitalLeaseObligation",
                      "CurrentCapitalLeaseObligation", "CurrentDebt", "OtherCurrentBorrowings", "LineOfCredit",
                      "CommercialPaper", "CurrentNotesPayable", "PensionandOtherPostRetirementBenefitPlansCurrent",
                      "CurrentProvisions", "PayablesAndAccruedExpenses", "CurrentAccruedExpenses", "InterestPayable",
                      "Payables", "OtherPayable", "DuetoRelatedPartiesCurrent", "DividendsPayable", "TotalTaxPayable",
                      "IncomeTaxPayable", "AccountsPayable", "TotalAssets", "TotalNonCurrentAssets",
                      "OtherNonCurrentAssets", "DefinedPensionBenefit", "NonCurrentPrepaidAssets",
                      "NonCurrentDeferredAssets", "NonCurrentDeferredTaxesAssets", "DuefromRelatedPartiesNonCurrent",
                      "NonCurrentNoteReceivables", "NonCurrentAccountsReceivable", "FinancialAssets",
                      "InvestmentsAndAdvances", "OtherInvestments", "InvestmentinFinancialAssets",
                      "HeldToMaturitySecurities", "AvailableForSaleSecurities",
                      "FinancialAssetsDesignatedasFairValueThroughProfitorLossTotal", "TradingSecurities",
                      "LongTermEquityInvestment", "InvestmentsinJointVenturesatCost",
                      "InvestmentsInOtherVenturesUnderEquityMethod", "InvestmentsinAssociatesatCost",
                      "InvestmentsinSubsidiariesatCost", "InvestmentProperties", "GoodwillAndOtherIntangibleAssets",
                      "OtherIntangibleAssets", "Goodwill", "NetPPE", "AccumulatedDepreciation", "GrossPPE", "Leases",
                      "ConstructionInProgress", "OtherProperties", "MachineryFurnitureEquipment",
                      "BuildingsAndImprovements", "LandAndImprovements", "Properties", "CurrentAssets",
                      "OtherCurrentAssets", "HedgingAssetsCurrent", "AssetsHeldForSaleCurrent", "CurrentDeferredAssets",
                      "CurrentDeferredTaxesAssets", "RestrictedCash", "PrepaidAssets", "Inventory",
                      "InventoriesAdjustmentsAllowances", "OtherInventories", "FinishedGoods", "WorkInProcess",
                      "RawMaterials", "Receivables", "ReceivablesAdjustmentsAllowances", "OtherReceivables",
                      "DuefromRelatedPartiesCurrent", "TaxesReceivable", "AccruedInterestReceivable", "NotesReceivable",
                      "LoansReceivable", "AccountsReceivable", "AllowanceForDoubtfulAccountsReceivable",
                      "GrossAccountsReceivable", "CashCashEquivalentsAndShortTermInvestments",
                      "OtherShortTermInvestments", "CashAndCashEquivalents", "CashEquivalents", "CashFinancial",
                      "CashCashEquivalentsAndFederalFundsSold"],
    'cash-flow': ["ForeignSales", "DomesticSales", "AdjustedGeographySegmentData", "FreeCashFlow",
                  "RepurchaseOfCapitalStock", "RepaymentOfDebt", "IssuanceOfDebt", "IssuanceOfCapitalStock",
                  "CapitalExpenditure", "InterestPaidSupplementalData", "IncomeTaxPaidSupplementalData",
                  "EndCashPosition", "OtherCashAdjustmentOutsideChangeinCash", "BeginningCashPosition",
                  "EffectOfExchangeRateChanges", "ChangesInCash", "OtherCashAdjustmentInsideChangeinCash",
                  "CashFlowFromDiscontinuedOperation", "FinancingCashFlow", "CashFromDiscontinuedFinancingActivities",
                  "CashFlowFromContinuingFinancingActivities", "NetOtherFinancingCharges", "InterestPaidCFF",
                  "ProceedsFromStockOptionExercised", "CashDividendsPaid", "PreferredStockDividendPaid",
                  "CommonStockDividendPaid", "NetPreferredStockIssuance", "PreferredStockPayments",
                  "PreferredStockIssuance", "NetCommonStockIssuance", "CommonStockPayments", "CommonStockIssuance",
                  "NetIssuancePaymentsOfDebt", "NetShortTermDebtIssuance", "ShortTermDebtPayments",
                  "ShortTermDebtIssuance", "NetLongTermDebtIssuance", "LongTermDebtPayments", "LongTermDebtIssuance",
                  "InvestingCashFlow", "CashFromDiscontinuedInvestingActivities",
                  "CashFlowFromContinuingInvestingActivities", "NetOtherInvestingChanges", "InterestReceivedCFI",
                  "DividendsReceivedCFI", "NetInvestmentPurchaseAndSale", "SaleOfInvestment", "PurchaseOfInvestment",
                  "NetInvestmentPropertiesPurchaseAndSale", "SaleOfInvestmentProperties",
                  "PurchaseOfInvestmentProperties", "NetBusinessPurchaseAndSale", "SaleOfBusiness",
                  "PurchaseOfBusiness", "NetIntangiblesPurchaseAndSale", "SaleOfIntangibles", "PurchaseOfIntangibles",
                  "NetPPEPurchaseAndSale", "SaleOfPPE", "PurchaseOfPPE", "CapitalExpenditureReported",
                  "OperatingCashFlow", "CashFromDiscontinuedOperatingActivities",
                  "CashFlowFromContinuingOperatingActivities", "TaxesRefundPaid", "InterestReceivedCFO",
                  "InterestPaidCFO", "DividendReceivedCFO", "DividendPaidCFO", "ChangeInWorkingCapital",
                  "ChangeInOtherWorkingCapital", "ChangeInOtherCurrentLiabilities", "ChangeInOtherCurrentAssets",
                  "ChangeInPayablesAndAccruedExpense", "ChangeInAccruedExpense", "ChangeInInterestPayable",
                  "ChangeInPayable", "ChangeInDividendPayable", "ChangeInAccountPayable", "ChangeInTaxPayable",
                  "ChangeInIncomeTaxPayable", "ChangeInPrepaidAssets", "ChangeInInventory", "ChangeInReceivables",
                  "ChangesInAccountReceivables", "OtherNonCashItems", "ExcessTaxBenefitFromStockBasedCompensation",
                  "StockBasedCompensation", "UnrealizedGainLossOnInvestmentSecurities", "ProvisionandWriteOffofAssets",
                  "AssetImpairmentCharge", "AmortizationOfSecurities", "DeferredTax", "DeferredIncomeTax",
                  "DepreciationAmortizationDepletion", "Depletion", "DepreciationAndAmortization",
                  "AmortizationCashFlow", "AmortizationOfIntangibles", "Depreciation", "OperatingGainsLosses",
                  "PensionAndEmployeeBenefitExpense", "EarningsLossesFromEquityInvestments",
                  "GainLossOnInvestmentSecurities", "NetForeignCurrencyExchangeGainLoss", "GainLossOnSaleOfPPE",
                  "GainLossOnSaleOfBusiness", "NetIncomeFromContinuingOperations",
                  "CashFlowsfromusedinOperatingActivitiesDirect", "TaxesRefundPaidDirect", "InterestReceivedDirect",
                  "InterestPaidDirect", "DividendsReceivedDirect", "DividendsPaidDirect", "ClassesofCashPayments",
                  "OtherCashPaymentsfromOperatingActivities", "PaymentsonBehalfofEmployees",
                  "PaymentstoSuppliersforGoodsandServices", "ClassesofCashReceiptsfromOperatingActivities",
                  "OtherCashReceiptsfromOperatingActivities", "ReceiptsfromGovernmentGrants", "ReceiptsfromCustomers"]}

_PRICE_COLNAMES_ = ['Open', 'High', 'Low', 'Close', 'Adj Close']

quote_summary_valid_modules = (
    "summaryProfile",  # contains general information about the company
    "summaryDetail",  # prices + volume + market cap + etc
    "assetProfile",  # summaryProfile + company officers
    "fundProfile",
    "price",  # current prices
    "quoteType",  # quoteType
    "esgScores",  # Environmental, social, and governance (ESG) scores, sustainability and ethical performance of companies
    "incomeStatementHistory",
    "incomeStatementHistoryQuarterly",
    "balanceSheetHistory",
    "balanceSheetHistoryQuarterly",
    "cashFlowStatementHistory",
    "cashFlowStatementHistoryQuarterly",
    "defaultKeyStatistics",  # KPIs (PE, enterprise value, EPS, EBITA, and more)
    "financialData",  # Financial KPIs (revenue, gross margins, operating cash flow, free cash flow, and more)
    "calendarEvents",  # future earnings date
    "secFilings",  # SEC filings, such as 10K and 10Q reports
    "upgradeDowngradeHistory",  # upgrades and downgrades that analysts have given a company's stock
    "institutionOwnership",  # institutional ownership, holders and shares outstanding
    "fundOwnership",  # mutual fund ownership, holders and shares outstanding
    "majorDirectHolders",
    "majorHoldersBreakdown",
    "insiderTransactions",  # insider transactions, such as the number of shares bought and sold by company executives
    "insiderHolders",  # insider holders, such as the number of shares held by company executives
    "netSharePurchaseActivity",  # net share purchase activity, such as the number of shares bought and sold by company executives
    "earnings",  # earnings history
    "earningsHistory",
    "earningsTrend",  # earnings trend
    "industryTrend",
    "indexTrend",
    "sectorTrend",
    "recommendationTrend",
    "futuresChain",
)

# map last updated as of 2024.09.18
SECTOR_INDUSTY_MAPPING = {
    'basic-materials': {'specialty-chemicals',
                        'gold',
                        'building-materials',
                        'copper',
                        'steel',
                        'agricultural-inputs',
                        'chemicals',
                        'other-industrial-metals-mining',
                        'lumber-wood-production',
                        'aluminum',
                        'other-precious-metals-mining',
                        'coking-coal',
                        'paper-paper-products',
                        'silver'},
    'communication-services': {'internet-content-information',
                                'telecom-services',
                                'entertainment',
                                'electronic-gaming-multimedia',
                                'advertising-agencies',
                                'broadcasting',
                                'publishing'},
    'consumer-cyclical': {'internet-retail',
                            'auto-manufacturers',
                            'restaurants',
                            'home-improvement-retail',
                            'travel-services',
                            'specialty-retail',
                            'apparel-retail',
                            'residential-construction',
                            'footwear-accessories',
                            'packaging-containers',
                            'lodging',
                            'auto-parts',
                            'auto-truck-dealerships',
                            'gambling',
                            'resorts-casinos',
                            'leisure',
                            'apparel-manufacturing',
                            'personal-services',
                            'furnishings-fixtures-appliances',
                            'recreational-vehicles',
                            'luxury-goods',
                            'department-stores',
                            'textile-manufacturing'},
    'consumer-defensive': {'discount-stores',
                            'beverages-non-alcoholic',
                            'household-personal-products',
                            'packaged-foods',
                            'tobacco',
                            'confectioners',
                            'farm-products',
                            'food-distribution',
                            'grocery-stores',
                            'beverages-brewers',
                            'education-training-services',
                            'beverages-wineries-distilleries'},
    'energy': {'oil-gas-integrated',
                'oil-gas-midstream',
                'oil-gas-e-p',
                'oil-gas-equipment-services',
                'oil-gas-refining-marketing',
                'uranium',
                'oil-gas-drilling',
                'thermal-coal'},
    'financial-services': {'banks-diversified',
                            'credit-services',
                            'asset-management',
                            'insurance-diversified',
                            'banks-regional',
                            'capital-markets',
                            'financial-data-stock-exchanges',
                            'insurance-property-casualty',
                            'insurance-brokers',
                            'insurance-life',
                            'insurance-specialty',
                            'mortgage-finance',
                            'insurance-reinsurance',
                            'shell-companies',
                            'financial-conglomerates'},
    'healthcare': {'drug-manufacturers-general',
                    'healthcare-plans',
                    'biotechnology',
                    'medical-devices',
                    'diagnostics-research',
                    'medical-instruments-supplies',
                    'medical-care-facilities',
                    'drug-manufacturers-specialty-generic',
                    'health-information-services',
                    'medical-distribution',
                    'pharmaceutical-retailers'},
    'industrials': {'aerospace-defense',
                    'specialty-industrial-machinery',
                    'railroads',
                    'building-products-equipment',
                    'farm-heavy-construction-machinery',
                    'specialty-business-services',
                    'integrated-freight-logistics',
                    'waste-management',
                    'conglomerates',
                    'industrial-distribution',
                    'engineering-construction',
                    'rental-leasing-services',
                    'consulting-services',
                    'trucking',
                    'electrical-equipment-parts',
                    'airlines',
                    'tools-accessories',
                    'pollution-treatment-controls',
                    'security-protection-services',
                    'marine-shipping',
                    'metal-fabrication',
                    'infrastructure-operations',
                    'staffing-employment-services',
                    'airports-air-services',
                    'business-equipment-supplies'},
    'real-estate': {'reit-specialty',
                    'reit-industrial',
                    'reit-retail',
                    'reit-residential',
                    'reit-healthcare-facilities',
                    'real-estate-services',
                    'reit-office',
                    'reit-diversified',
                    'reit-mortgage',
                    'reit-hotel-motel',
                    'real-estate-development',
                    'real-estate-diversified'},
    'technology': {'software-infrastructure',
                    'semiconductors',
                    'consumer-electronics',
                    'software-application',
                    'information-technology-services',
                    'semiconductor-equipment-materials',
                    'communication-equipment',
                    'computer-hardware',
                    'electronic-components',
                    'scientific-technical-instruments',
                    'solar',
                    'electronics-computer-distribution'},
    'utilities': {'utilities-regulated-electric',
                    'utilities-renewable',
                    'utilities-diversified',
                    'utilities-regulated-gas',
                    'utilities-independent-power-producers',
                    'utilities-regulated-water'}
}
EQUITY_SCREENER_EQ_MAP = {
    "region": {
        "za", "ve", "vn", "us", "tw", "th", "tr", "sr", "sg", "sa", "se", "ru", "ro", "qa", "pt", "pk", "pl",
        "ph", "nz", "nl", "mx", "pe", "no", "my", "lv", "lt", "kw", "jp", "is", "il", "lk", "kr", "it", "in",
        "ie", "hu", "id", "hk", "gb", "fi", "eg", "dk", "gr", "fr", "es", "ee", "de", "cz", "cl", "ca", "be",
        "at", "cn", "br", "au", "ar", "ch"
    },
    "sector": {
        "Basic Materials", "Industrials", "Communication Services", "Healthcare",
        "Real Estate", "Technology", "Energy", "Utilities", "Financial Services",
        "Consumer Defensive", "Consumer Cyclical"
    },
    "exchanges": {
        "NMS", "NAS", "YHD", "NYQ", "NGM", "NCM", "BSE"
    },
    "peer_group": {
        "US Fund Equity Energy",
        "US CE Convertibles",
        "EAA CE UK Large-Cap Equity",
        "EAA CE Other",
        "US Fund Financial",
        "India CE Multi-Cap",
        "US Fund Foreign Large Blend",
        "US Fund Consumer Cyclical",
        "EAA Fund Global Equity Income",
        "China Fund Sector Equity Financial and Real Estate",
        "US Fund Equity Precious Metals",
        "EAA Fund RMB Bond - Onshore",
        "China Fund QDII Greater China Equity",
        "US Fund Large Growth",
        "EAA Fund Germany Equity",
        "EAA Fund Hong Kong Equity",
        "EAA CE UK Small-Cap Equity",
        "US Fund Natural Resources",
        "US CE Preferred Stock",
        "India Fund Sector - Financial Services",
        "US Fund Diversified Emerging Mkts",
        "EAA Fund South Africa & Namibia Equity",
        "China Fund QDII Sector Equity",
        "EAA CE Sector Equity Biotechnology",
        "EAA Fund Switzerland Equity",
        "US Fund Large Value",
        "EAA Fund Asia ex-Japan Equity",
        "US Fund Health",
        "US Fund China Region",
        "EAA Fund Emerging Europe ex-Russia Equity",
        "EAA Fund Sector Equity Industrial Materials",
        "EAA Fund Japan Large-Cap Equity",
        "EAA Fund EUR Corporate Bond",
        "US Fund Technology",
        "EAA CE Global Large-Cap Blend Equity",
        "Mexico Fund Mexico Equity",
        "US Fund Trading--Leveraged Equity",
        "EAA Fund Sector Equity Consumer Goods & Services",
        "US Fund Large Blend",
        "EAA Fund Global Flex-Cap Equity",
        "EAA Fund EUR Aggressive Allocation - Global",
        "EAA Fund China Equity",
        "EAA Fund Global Large-Cap Growth Equity",
        "US CE Options-based",
        "EAA Fund Sector Equity Financial Services",
        "EAA Fund Europe Large-Cap Blend Equity",
        "EAA Fund China Equity - A Shares",
        "EAA Fund USD Corporate Bond",
        "EAA Fund Eurozone Large-Cap Equity",
        "China Fund Aggressive Allocation Fund",
        "EAA Fund Sector Equity Technology",
        "EAA Fund Global Emerging Markets Equity",
        "EAA Fund EUR Moderate Allocation - Global",
        "EAA Fund Other Bond",
        "EAA Fund Denmark Equity",
        "EAA Fund US Large-Cap Blend Equity",
        "India Fund Large-Cap",
        "Paper & Forestry",
        "Containers & Packaging",
        "US Fund Miscellaneous Region",
        "Energy Services",
        "EAA Fund Other Equity",
        "Homebuilders",
        "Construction Materials",
        "China Fund Equity Funds",
        "Steel",
        "Consumer Durables",
        "EAA Fund Global Large-Cap Blend Equity",
        "Transportation Infrastructure",
        "Precious Metals",
        "Building Products",
        "Traders & Distributors",
        "Electrical Equipment",
        "Auto Components",
        "Construction & Engineering",
        "Aerospace & Defense",
        "Refiners & Pipelines",
        "Diversified Metals",
        "Textiles & Apparel",
        "Industrial Conglomerates",
        "Household Products",
        "Commercial Services",
        "Food Retailers",
        "Semiconductors",
        "Media",
        "Automobiles",
        "Consumer Services",
        "Technology Hardware",
        "Transportation",
        "Telecommunication Services",
        "Oil & Gas Producers",
        "Machinery",
        "Retailing",
        "Healthcare",
        "Chemicals",
        "Food Products",
        "Diversified Financials",
        "Real Estate",
        "Insurance",
        "Utilities",
        "Pharmaceuticals",
        "Software & Services",
        "Banks"
    }
}
EQUITY_SCREENER_FIELDS = {
    # EQ Fields
    "region",
    "sector",
    "peer_group",
    "exchanges",

    # price
    "eodprice",
    "intradaypricechange",
    "lastclosemarketcap.lasttwelvemonths",
    "percentchange",
    "lastclose52weekhigh.lasttwelvemonths",
    "fiftytwowkpercentchange",
    "intradayprice",
    "lastclose52weeklow.lasttwelvemonths",
    "intradaymarketcap",

    # trading
    "beta",
    "avgdailyvol3m",
    "pctheldinsider",
    "pctheldinst",
    "dayvolume",
    "eodvolume",

    # short interest
    "short_percentage_of_shares_outstanding.value",
    "short_interest.value",
    "short_percentage_of_float.value",
    "days_to_cover_short.value",
    "short_interest_percentage_change.value",

    # valuation
    "bookvalueshare.lasttwelvemonths",
    "lastclosemarketcaptotalrevenue.lasttwelvemonths",
    "lastclosetevtotalrevenue.lasttwelvemonths",
    "pricebookratio.quarterly",
    "peratio.lasttwelvemonths",
    "lastclosepricetangiblebookvalue.lasttwelvemonths",
    "lastclosepriceearnings.lasttwelvemonths",
    "pegratio_5y",

    # profitability
    "consecutive_years_of_dividend_growth_count",
    "returnonassets.lasttwelvemonths",
    "returnonequity.lasttwelvemonths",
    "forward_dividend_per_share",
    "forward_dividend_yield",
    "returnontotalcapital.lasttwelvemonths",

    # leverage
    "lastclosetevebit.lasttwelvemonths",
    "netdebtebitda.lasttwelvemonths",
    "totaldebtequity.lasttwelvemonths",
    "ltdebtequity.lasttwelvemonths",
    "ebitinterestexpense.lasttwelvemonths",
    "ebitdainterestexpense.lasttwelvemonths",
    "lastclosetevebitda.lasttwelvemonths",
    "totaldebtebitda.lasttwelvemonths",

    # liquidity
    "quickratio.lasttwelvemonths",
    "altmanzscoreusingtheaveragestockinformationforaperiod.lasttwelvemonths",
    "currentratio.lasttwelvemonths",
    "operatingcashflowtocurrentliabilities.lasttwelvemonths",

    # income statement
    "totalrevenues.lasttwelvemonths",
    "netincomemargin.lasttwelvemonths",
    "grossprofit.lasttwelvemonths",
    "ebitda1yrgrowth.lasttwelvemonths",
    "dilutedepscontinuingoperations.lasttwelvemonths",
    "quarterlyrevenuegrowth.quarterly",
    "epsgrowth.lasttwelvemonths",
    "netincomeis.lasttwelvemonths",
    "ebitda.lasttwelvemonths",
    "dilutedeps1yrgrowth.lasttwelvemonths",
    "totalrevenues1yrgrowth.lasttwelvemonths",
    "operatingincome.lasttwelvemonths",
    "netincome1yrgrowth.lasttwelvemonths",
    "grossprofitmargin.lasttwelvemonths",
    "ebitdamargin.lasttwelvemonths",
    "ebit.lasttwelvemonths",
    "basicepscontinuingoperations.lasttwelvemonths",
    "netepsbasic.lasttwelvemonths"
    "netepsdiluted.lasttwelvemonths",

    # balance sheet
    "totalassets.lasttwelvemonths",
    "totalcommonsharesoutstanding.lasttwelvemonths",
    "totaldebt.lasttwelvemonths",
    "totalequity.lasttwelvemonths",
    "totalcurrentassets.lasttwelvemonths",
    "totalcashandshortterminvestments.lasttwelvemonths",
    "totalcommonequity.lasttwelvemonths",
    "totalcurrentliabilities.lasttwelvemonths",
    "totalsharesoutstanding",

    # cash flow
    "forward_dividend_yield",
    "leveredfreecashflow.lasttwelvemonths",
    "capitalexpenditure.lasttwelvemonths",
    "cashfromoperations.lasttwelvemonths",
    "leveredfreecashflow1yrgrowth.lasttwelvemonths",
    "unleveredfreecashflow.lasttwelvemonths",
    "cashfromoperations1yrgrowth.lasttwelvemonths",

    # ESG
    "esg_score",
    "environmental_score",
    "governance_score",
    "social_score",
    "highest_controversy"
}

PREDEFINED_SCREENER_BODY_MAP = {
    'aggressive_small_caps': {"offset":0,"size":25,"sortField":"eodvolume","sortType":"desc","quoteType":"equity","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NMS"]},{"operator":"eq","operands":["exchange","NYQ"]}]},{"operator":"or","operands":[{"operator":"LT","operands":["epsgrowth.lasttwelvemonths",15]}]}]},"userId":"","userIdType":"guid"},
    'day_gainers': {"offset":0,"size":25,"sortField":"percentchange","sortType":"DESC","quoteType":"EQUITY","query":{"operator":"AND","operands":[{"operator":"gt","operands":["percentchange",3]},{"operator":"eq","operands":["region","us"]},{"operator":"or","operands":[{"operator":"BTWN","operands":["intradaymarketcap",2000000000,10000000000]},{"operator":"BTWN","operands":["intradaymarketcap",10000000000,100000000000]},{"operator":"GT","operands":["intradaymarketcap",100000000000]}]},{"operator":"gte","operands":["intradayprice",5]},{"operator":"gt","operands":["dayvolume",15000]}]},"userId":"","userIdType":"guid"},
    'day_losers': {"offset":0,"size":25,"sortField":"percentchange","sortType":"ASC","quoteType":"EQUITY","query":{"operator":"AND","operands":[{"operator":"lt","operands":["percentchange",-2.5]},{"operator":"eq","operands":["region","us"]},{"operator":"or","operands":[{"operator":"BTWN","operands":["intradaymarketcap",2000000000,10000000000]},{"operator":"BTWN","operands":["intradaymarketcap",10000000000,100000000000]},{"operator":"GT","operands":["intradaymarketcap",100000000000]}]},{"operator":"gte","operands":["intradayprice",5]},{"operator":"gt","operands":["dayvolume",20000]}]},"userId":"","userIdType":"guid"},
    'growth_technology_stocks': {"offset":0,"size":25,"sortField":"eodvolume","sortType":"desc","quoteType":"equity","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"BTWN","operands":["quarterlyrevenuegrowth.quarterly",50,100]},{"operator":"GT","operands":["quarterlyrevenuegrowth.quarterly",100]},{"operator":"BTWN","operands":["quarterlyrevenuegrowth.quarterly",25,50]}]},{"operator":"or","operands":[{"operator":"BTWN","operands":["epsgrowth.lasttwelvemonths",25,50]},{"operator":"BTWN","operands":["epsgrowth.lasttwelvemonths",50,100]},{"operator":"GT","operands":["epsgrowth.lasttwelvemonths",100]}]},{"operator":"eq","operands":["sector","Technology"]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NMS"]},{"operator":"eq","operands":["exchange","NYQ"]}]}]},"userId":"","userIdType":"guid"},
    'most_actives': {"offset":0,"size":25,"sortField":"dayvolume","sortType":"DESC","quoteType":"EQUITY","query":{"operator":"AND","operands":[{"operator":"eq","operands":["region","us"]},{"operator":"or","operands":[{"operator":"BTWN","operands":["intradaymarketcap",10000000000,100000000000]},{"operator":"GT","operands":["intradaymarketcap",100000000000]},{"operator":"BTWN","operands":["intradaymarketcap",2000000000,10000000000]}]},{"operator":"gt","operands":["dayvolume",5000000]}]},"userId":"","userIdType":"guid"},
    'most_shorted_stocks': {"size":25,"offset":0,"sortField":"short_percentage_of_shares_outstanding.value","sortType":"DESC","quoteType":"EQUITY","topOperator":"AND","query":{"operator":"AND","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["region","us"]}]},{"operator":"gt","operands":["intradayprice",1]},{"operator":"gt","operands":["avgdailyvol3m",200000]}]},"userId":"","userIdType":"guid"},
    'small_cap_gainers': {"offset":0,"size":25,"sortField":"eodvolume","sortType":"desc","quoteType":"equity","query":{"operator":"and","operands":[{"operator":"lt","operands":["intradaymarketcap",2000000000]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NMS"]},{"operator":"eq","operands":["exchange","NYQ"]}]}]},"userId":"","userIdType":"guid"},
    'undervalued_growth_stocks': {"offset":0,"size":25,"sortType":"DESC","sortField":"eodvolume","quoteType":"EQUITY","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"BTWN","operands":["peratio.lasttwelvemonths",0,20]}]},{"operator":"or","operands":[{"operator":"LT","operands":["pegratio_5y",1]}]},{"operator":"or","operands":[{"operator":"BTWN","operands":["epsgrowth.lasttwelvemonths",25,50]},{"operator":"BTWN","operands":["epsgrowth.lasttwelvemonths",50,100]},{"operator":"GT","operands":["epsgrowth.lasttwelvemonths",100]}]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NMS"]},{"operator":"eq","operands":["exchange","NYQ"]}]}]},"userId":"","userIdType":"guid"},
    'undervalued_large_caps': {"offset":0,"size":25,"sortField":"eodvolume","sortType":"desc","quoteType":"equity","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"BTWN","operands":["peratio.lasttwelvemonths",0,20]}]},{"operator":"lt","operands":["pegratio_5y",1]},{"operator":"btwn","operands":["intradaymarketcap",10000000000,100000000000]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NMS"]},{"operator":"eq","operands":["exchange","NYQ"]}]}]},"userId":"","userIdType":"guid"},
    'conservative_foreign_funds': {"offset":0,"size":25,"sortType":"DESC","sortField":"fundnetassets","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["categoryname","Foreign Large Value"]},{"operator":"EQ","operands":["categoryname","Foreign Large Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Large Growth"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Growth"]},{"operator":"EQ","operands":["categoryname","Foreign Large Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Value"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Value"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Value"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Blend"]},{"operator":"EQ","operands":["categoryname","Foreign Small/Mid Value"]}]},{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",4]},{"operator":"EQ","operands":["performanceratingoverall",5]}]},{"operator":"lt","operands":["initialinvestment",100001]},{"operator":"lt","operands":["annualreturnnavy1categoryrank",50]},{"operator":"or","operands":[{"operator":"EQ","operands":["riskratingoverall",1]},{"operator":"EQ","operands":["riskratingoverall",3]},{"operator":"EQ","operands":["riskratingoverall",2]}]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"},
    'high_yield_bond': {"offset":0,"size":25,"sortType":"DESC","sortField":"fundnetassets","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",4]},{"operator":"EQ","operands":["performanceratingoverall",5]}]},{"operator":"lt","operands":["initialinvestment",100001]},{"operator":"lt","operands":["annualreturnnavy1categoryrank",50]},{"operator":"or","operands":[{"operator":"EQ","operands":["riskratingoverall",1]},{"operator":"EQ","operands":["riskratingoverall",3]},{"operator":"EQ","operands":["riskratingoverall",2]}]},{"operator":"or","operands":[{"operator":"EQ","operands":["categoryname","High Yield Bond"]}]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"},
    'portfolio_anchors': {"offset":0,"size":25,"sortType":"DESC","sortField":"fundnetassets","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["categoryname","Large Blend"]}]},{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",4]},{"operator":"EQ","operands":["performanceratingoverall",5]}]},{"operator":"lt","operands":["initialinvestment",100001]},{"operator":"lt","operands":["annualreturnnavy1categoryrank",50]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"},
    'solid_large_growth_funds': {"offset":0,"size":25,"sortType":"DESC","sortField":"fundnetassets","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["categoryname","Large Growth"]}]},{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",5]},{"operator":"EQ","operands":["performanceratingoverall",4]}]},{"operator":"lt","operands":["initialinvestment",100001]},{"operator":"lt","operands":["annualreturnnavy1categoryrank",50]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"},
    'solid_midcap_growth_funds': {"offset":0,"size":25,"sortType":"DESC","sortField":"fundnetassets","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"or","operands":[{"operator":"EQ","operands":["categoryname","Mid-Cap Growth"]}]},{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",5]},{"operator":"EQ","operands":["performanceratingoverall",4]}]},{"operator":"lt","operands":["initialinvestment",100001]},{"operator":"lt","operands":["annualreturnnavy1categoryrank",50]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"},
    'top_mutual_funds': {"offset":0,"size":25,"sortType":"DESC","sortField":"percentchange","quoteType":"MUTUALFUND","query":{"operator":"and","operands":[{"operator":"gt","operands":["intradayprice",15]},{"operator":"or","operands":[{"operator":"EQ","operands":["performanceratingoverall",5]},{"operator":"EQ","operands":["performanceratingoverall",4]}]},{"operator":"gt","operands":["initialinvestment",1000]},{"operator":"or","operands":[{"operator":"eq","operands":["exchange","NAS"]}]}]},"userId":"","userIdType":"guid"}
}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yfinance - market data downloader
# https://github.com/ranaroussi/yfinance
#
# Copyright 2017-2019 Ran Aroussi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from . import version
from .ticker import Ticker
from .tickers import Tickers
from .multi import download
from .utils import enable_debug_mode
from .cache import set_tz_cache_location
from .domain.sector import Sector
from .domain.industry import Industry
from .screener.screener import Screener
from .screener.screener_query import EquityQuery

__version__ = version.version
__author__ = "Ran Aroussi"

import warnings
warnings.filterwarnings('default', category=DeprecationWarning, module='^yfinance')

__all__ = ['download', 'Ticker', 'Tickers', 'enable_debug_mode', 'set_tz_cache_location', 'Sector', 'Industry',
           'EquityQuery','Screener']



import numbers
from typing import List, Union, Dict, Set

from yfinance.const import EQUITY_SCREENER_EQ_MAP, EQUITY_SCREENER_FIELDS
from yfinance.exceptions import YFNotImplementedError

class Query:
    def __init__(self, operator: str, operand: Union[numbers.Real, str, List['Query']]):
        self.operator = operator
        self.operands = operand
    
    def to_dict(self) -> Dict:
        raise YFNotImplementedError('to_dict() needs to be implemented by children classes')

class EquityQuery(Query):
    def __init__(self, operator: str, operand: Union[numbers.Real, str, List['EquityQuery']]):
        operator = operator.upper()

        if not isinstance(operand, list):
            raise TypeError('Invalid operand type')
        if len(operand) <= 0:
            raise ValueError('Invalid field for Screener')
            
        if operator in {'OR','AND'}: 
            self._validate_or_and_operand(operand)
        elif operator == 'EQ': 
            self._validate_eq_operand(operand)
        elif operator == 'BTWN': 
            self._validate_btwn_operand(operand)
        elif operator in {'GT','LT'}: 
            self._validate_gt_lt(operand)
        else: 
            raise ValueError('Invalid Operator Value')

        self.operator = operator
        self.operands = operand
        self._valid_eq_map = EQUITY_SCREENER_EQ_MAP
        self._valid_fields = EQUITY_SCREENER_FIELDS
        
    @property
    def valid_eq_map(self) -> Dict:
        return self._valid_eq_map

    @property
    def valid_fields(self) -> Set:
        return self._valid_fields
    
    def _validate_or_and_operand(self, operand: List['EquityQuery']) -> None:
        if len(operand) <= 1: 
            raise ValueError('Operand must be length longer than 1')
        if all(isinstance(e, EquityQuery) for e in operand) is False: 
            raise TypeError('Operand must be type EquityQuery for OR/AND')

    def _validate_eq_operand(self, operand: List[Union[str, numbers.Real]]) -> None:
        if len(operand) != 2:
            raise ValueError('Operand must be length 2 for EQ')
        if operand[0] not in EQUITY_SCREENER_FIELDS:
            raise ValueError('Invalid field for Screener')
        if operand[0] not in EQUITY_SCREENER_EQ_MAP:
            raise ValueError('Invalid EQ key')
        if operand[1] not in EQUITY_SCREENER_EQ_MAP[operand[0]]:
            raise ValueError('Invalid EQ value')
    
    def _validate_btwn_operand(self, operand: List[Union[str, numbers.Real]]) -> None:
        if len(operand) != 3: 
            raise ValueError('Operand must be length 3 for BTWN')
        if operand[0] not in EQUITY_SCREENER_FIELDS:
            raise ValueError('Invalid field for Screener')
        if isinstance(operand[1], numbers.Real) is False:
            raise TypeError('Invalid comparison type for BTWN')
        if isinstance(operand[2], numbers.Real) is False:
            raise TypeError('Invalid comparison type for BTWN')

    def _validate_gt_lt(self, operand: List[Union[str, numbers.Real]]) -> None:
        if len(operand) != 2:
            raise ValueError('Operand must be length 2 for GT/LT')
        if operand[0] not in EQUITY_SCREENER_FIELDS:
            raise ValueError('Invalid field for Screener')
        if isinstance(operand[1], numbers.Real) is False:
            raise TypeError('Invalid comparison type for GT/LT')

    def to_dict(self) -> Dict:
        return {
            "operator": self.operator,
            "operands": [operand.to_dict() if isinstance(operand, EquityQuery) else operand for operand in self.operands]
        }